/** 
 *  Yudit Unicode Editor Source File
 *
 *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
 *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
 *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2,
 *  dated June 1991. See file COPYYING for details.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef SGlyph_h
#define SGlyph_h

#include "stoolkit/SVector.h"
#include "stoolkit/STypes.h"
#include "stoolkit/SUniMap.h"
#include "stoolkit/SString.h"
#include "stoolkit/SCharClass.h"
#include "swindow/SColor.h"


class SGlyphShared
{
public:
  char    cluster;
  char    composing;
  char    type;
  bool    lineend;
  bool    tab;
  bool    shaped;

  SS_UCS4 precomposed;
  SS_UCS4 mirror;
  /* 
   * This is ucs4v:
   *
   * For cluster == 0, precomposed is precomposed character if any.
   * shape[4],decomposition,extra-composed
   * ^                      ^
   * |                      +-- composing, or 0 
   * |                          
   * +------ if shaped it has 4 chars here. It can be private char.
   * 
   * For cluster == n, precomposed is cluster private charID if any
   * unicode-cluster,memory-cluster,extra-composed
   *                 ^              ^
   *                 |              +-------composing or 0
   *                 |
   *                 +---cluster-offset
   */
  SV_UCS4 ucs4v;
};

SGlyphShared* getGlyphShared (SS_UCS4 c);

class SGlyph : public SObject
{
public:
  enum SType {
       CTRL = 0,
       LF = '\n',
       CR = '\r',
       TAB = '\t',
       RLO = 0x202e, /* right-to-left override */
       LRO = 0x202d,/* leftto-right override */
       PDF = 0x202c,  /* pop directional format */
       PS = 0x2029, /* paragraph separator */
       LS = 0x2028 /* line separator */
  };

  enum SShape {
      ISOLATED=0,
      INITIAL=1,
      MEDIAL=2,
      FINAL=3,
      /* Syriac U+0710 extra shapes */
      SYRIAC_ALAPH_FINAL_FJ=4,
      SYRIAC_ALAPH_ISOLATED_FN=5,
      SYRIAC_ALAPH_ISOLATED_FX=6,
      NOSHAPE=SD_NOSHAPE
  };

  enum SLineEnd {
      ELF=0,
      ECR,
      ECRLF,
      EPS,
      ELS
  };

  SGlyph (SD_CharClass type, const SV_UCS4 &decomp, SS_UCS4 comp, 
          SS_UCS4 mirror, bool shaped, 
          unsigned int clusterindex, 
          unsigned int composingIndex); 

  SGlyph (const SGlyph& glyph);
  SGlyph (SD_CharClass type, SGlyphShared* _shared);

  SGlyph operator=(const SGlyph& glyph);
  virtual ~SGlyph ();
  virtual SObject*  clone() const;

  bool operator == (const SGlyph& g2) const;
  bool operator != (const SGlyph& g2) const;

  bool operator == (const SType& type) const;
  bool operator != (const SType& type) const;

  bool isWhiteSpace() const;
  bool isDelimiter() const;
  bool isNumber() const;
  bool isTransparent() const;
  bool isEOP() const;
  bool isEOL() const;
  bool isTab() const;

  bool isCluster() const;
  bool isYuditLigature() const;
  bool isYuditComposition() const;
  bool isSpecial()  const;
  bool isMirrorable() const;
  SD_CharClass getType()  const;

  SString charKey() const;

  bool setShape (const SGlyph* before, const SGlyph* after);

  
  /* Work on characters */
  const SS_UCS4 getChar() const;
  const SS_UCS4 getMirroredChar() const;
  const SS_UCS4 getShapedChar() const;
  const SS_UCS4 getFirstChar () const;

  /* Work on decompositions */
  SV_UCS4 getChars() const; /* returns decomposed or precomposed  */

  unsigned int decompSize() const;
  unsigned int compSize() const;

  const SS_UCS4* getShapeArray() const; /* fast */
  const SS_UCS4* getDecompArray() const; /* fast */
  const SS_UCS4* getCompArray() const; /* fast */
  const SS_UCS4* getShapeFallback() const; /* fast */

  /* array operator works on memory represenation */
  SS_UCS4 operator[] (unsigned int index) const; /* fast */

  /* now comes the real hack. */
  bool addComposing(SS_UCS4 c);
  SS_UCS4 removeComposing();

  /* These should be constants */
  bool    underlined;
  bool    selected;
  bool    usePrecomp;
  bool    lr;
  char    currentShape;


private:
  char getShape (const SGlyph* before, const SGlyph* after);
  SGlyphShared * shared;

};

void addFallbackShapes (SUniMap* shaper, const SS_UCS4* shapes,
  const SS_UCS4* chars, unsigned int size);

#endif /* SGlyph_h */
