/** 
 *  Yudit Unicode Editor Source File
 *
 *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
 *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
 *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2,
 *  dated June 1991. See file COPYYING for details.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "stoolkit/SParagraph.h"
#include "stoolkit/SUniMap.h"
#include "stoolkit/SCluster.h"

#define SD_MAX_COMPOSE 100

static unsigned int 
split(const SV_UCS4& ucs4, SParagraph* gl, unsigned int from);

SParagraph::SParagraph(void)
{
}

/**
 * Lift off one paragraph from vector starting from start
 * at the end the line size is zero.
 */
SParagraph::SParagraph(const SV_UCS4& buffer, unsigned int *index)
{
  unsigned int start = *index;
  unsigned int end = start;

  /* split the lines and do stoolkit composition */
  end = split (buffer, this, start);
  /* repeat while Line Separator */
  while (size() > 0 && (peek(size()-1))->isEOL() && start != end)
  {
    start = end;
    end = split (buffer, this, start);
  }
  *index = end;
}

SParagraph::~SParagraph()
{
}

/**
 * try to shape a single segment.
 * return true if it is possible.
 * @param before is the visual glyph before this line.
 * @param after is the visual glyph after this line.
 * Todo: do glyphline-breaking of shaping is not possible, use
 * lineShaper to get glyph breakdown.
 */
void
SParagraph::reShapeSegment ( const SGlyph* _before, const SGlyph* _after)
{
  unsigned int gsize = size();
  const SGlyph* before = _before;
  const SGlyph* after = _after;
  for (unsigned int i=0; i<gsize; i++)
  {
    SGlyph* g = (SGlyph*) peek(i);
    if (g->getShapeArray()==0) continue;

    int cbefore = getNonTransparentBefore (i);
    int cafter = getNonTransparentAfter (i);
    before = (cbefore >= 0) ? peek((unsigned int)cbefore) : _before;
    after = (cafter >= 0) ? peek((unsigned int)cafter) : _after;
    g->setShape (before, after);
  }
  return;
}

/**
 * Get the first non-spacing glyph left visual order in array, 
 * starting here and incuding this.
 * return the index or -1 if not found.
 */
int 
SParagraph::getNonTransparentBefore(unsigned int index)
{
 if (index > size()) return -1;
 for (int i=((int)index) - 1; i >= 0; i--)
 {
   const SGlyph* g = peek((unsigned int)i);
   if (!g->isTransparent()) return i; 
 }
 return -1;
}

/**
 * Get the first non-spacing glyph right, in visual order, 
 * starting here and including this.
 * return the index or -1 if not found.
 */
int 
SParagraph::getNonTransparentAfter(unsigned int index)
{
 for (int i=index +1; i < (int)size(); i++)
 {
   const SGlyph* g = peek((unsigned int)i);
   if (!g->isTransparent()) return i; 
 }
 return -1;
}

/**
 * Split a text into lines
 * @param ucs4 is a text that can contain newline chars.
 * @param from is the starting index.
 * @parem gl is the return glyphline.
 * @return ending index, that is equal to starting index 
 * if ther is no more data.
 */
static unsigned int
split (const SV_UCS4& ucs4, SParagraph* gl, unsigned int from)
{
  SUniMap composer ("precompose");
  SUniMap category ("category");
  SUniMap shaper ("shape");
  SUniMap mirroring ("mirroring");

  gl->clear();
  unsigned int i=from;
  SS_UCS4 composition;
  SS_UCS4 mirror;
  bool    usePrecomposed;
  bool    isShaped;
  bool    isLigature;
  bool    isLineEnd;
  //gl->ensure (ucs4.size());
  SGlyphShared* shr = 0;
  unsigned int compIndex = 0;
  while (i<ucs4.size())
  {
    SD_CharClass chartype = SD_CC_Xx;
    unsigned int n =0;
    SV_UCS4 ret;
    composition = 0;
    compIndex = 0;
    mirror = 0;
    usePrecomposed = false;
    isShaped = false;
    isLigature = false;
    isLineEnd = true;
    unsigned int clusterIndex = 0;
    SV_UCS4 du4;
    shr = 0;
    switch (ucs4[i])
    {
    case SGlyph::CR:
      du4.append (ucs4[i++]);
      if (i<ucs4.size() && ucs4[i] == SGlyph::LF)
      {
        du4.append (ucs4[i++]);
      }
      break;
    case SGlyph::LF:
      du4.append (ucs4[i++]);
      if (i<ucs4.size() && ucs4[i] == SGlyph::CR)
      {
        du4.append (ucs4[i++]);
      }
      break;
    case SGlyph::PS:
    case SGlyph::LS:
    case SGlyph::RLO:
    case SGlyph::LRO:
    case SGlyph::PDF:
      du4.append (ucs4[i++]);
      break ;
    default:
      isLineEnd = false;
      /* shaping stand alone characters is allowed. */
      if (category.isOK())
      {
        /* max 0x1e */
        chartype = (SD_CharClass) category.encode (ucs4[i]);
      }
      if (mirroring.isOK())
      {
        mirror = mirroring.encode (ucs4[i]);
      }
      
      if (shaper.isOK())
      {
        /* encode shapes if possible */
        unsigned int n = shaper.lift (ucs4, i, false, &ret);
        /* the composition comes at the end  - if any */
        if (n>=i+1 && ret.size()==4)
        {
          composition =  ucs4[i];
          /* The four variants for shaping */
          du4.append (ret[0]);
          du4.append (ret[1]);
          du4.append (ret[2]);
          du4.append (ret[3]);
          if (n>i+1) // composed of several
          {
            usePrecomposed = false;
            composition = 0;
            /* append the extra bits at the end. */
            for (unsigned int j=i; j<n; j++) du4.append (ucs4[j]);

            /* this will make sure we will have something to 
               display if any of the shapes will fail to render */
            addFallbackShapes (&shaper, du4.array(), 
                &du4.array()[4], n-i);
          }
          else /* n == i+1: only one character read */
          {
            usePrecomposed = true;
            /* we still have some extra characters. Is it a composition? */
            if (i+1<ucs4.size())
            {
              SV_UCS4 retc;
              /* good to have - for cacheing ...*/
              unsigned int np = composer.lift (ucs4, i, false, &retc);
              if (np>i+1 && retc.size() == 1)
              {
                composition = retc[0];
                usePrecomposed = false;
                ret.clear();
                /* replace the shapes for this newly composed glyph */
                unsigned int ns = shaper.lift (retc, 0, false, &ret);
                if (ns==1 && ret.size()==4)
                {
                  du4.replace (0, ret[0]);
                  du4.replace (1, ret[1]);
                  du4.replace (2, ret[2]);
                  du4.replace (3, ret[3]);
                }
                /* the composition comes at the end. */
                for (unsigned int j=i; j<np; j++) du4.append (ucs4[j]);
                n = np;
              }
            }
             /*  decompositions, if any */ 
            if (usePrecomposed) /* set if we decomposed this */
            {
              SV_UCS4 retc;
              unsigned int nd = composer.lift (ucs4, i, true, &retc);
              if (nd == i+1) /* fill in decomposition */
              {
                for (unsigned int j=0; j<retc.size(); j++) du4.append (retc[j]);
              }
            }
          }
          i = n;
          isShaped = true;
          ret.clear();
          break;
        }
        ret.clear();
      }
      /* check for clusters */
      if ((n = getCluster (ucs4, i, &ret)) > i
          && ret.size() > 0 && ret[0] < 0x80000000)
      {
         unsigned int j;
         for (j=0; j<ret.size(); j++)
         {
           if (ret[j] > 0x80000000 && ret[j] < 0xA0000000)
           {
              composition = ret[j];
              /* Save this unicode ligature. Fonts have unicode order.  */
              putLigatureCluster (composition, du4.array(), du4.size());
              putLigatureUnicode (composition, &ucs4.array()[i], n-i);
              break;
           }
           du4.append (ret[j]);
         }
         clusterIndex = j;
         if (i>=n) du4.append (ucs4[i++]); /* fallback - error */
         while (i<n)
         {
          du4.append (ucs4[i++]);
         }
         ret.clear ();
      }
      else if (composer.isOK () && category.isOK())
      {
        ret.clear();
        n = composer.lift (ucs4, i, false, &ret);
        if (ret.size()==1) /* it was  a composition */
        {
          composition = ret[0];
          if (i>=n) du4.append (ucs4[i++]); /* fallback - error */
          while (i<n)
          {
            du4.append (ucs4[i++]);
          }
          ret.clear();
        }
        else /* it was not a composition */
        {
          /* quick get on a precomposed character */
          shr = getGlyphShared (ucs4[i]);
          if (shr != 0)
          {
            /*
             * Glyphcache can have precomposed or single chars only. 
             * But these single chars have have some comps...
             */
            du4 = shr->ucs4v;
            /* we are saving memory in sglyph */
            if (du4.size()==0) du4.append (ucs4[i]);
            mirror = shr->mirror;
            composition = shr->precomposed; /* ucs4[i] */
            usePrecomposed = true;
            i++;
            break;
          }
          ret.clear();

          /* check if it can be broken down into a composition */
          unsigned int n = composer.lift (ucs4, i, true, &ret);

          if (ucs4[i] != 0 && n == i+1 
              && ret.size() > 0 && ret.size() <= SD_MAX_COMPOSE)
          {
            usePrecomposed = true;
            n = i+1;
            composition = ucs4[i++];
            du4.append (ret);
          }
          else
          {
            /* 
             * No composition, there is only a base. 
             * Extra compositions will be added later.
             */
            du4.append (ucs4[i++]);
          }
        }
      }
      else
      {
        du4.append (ucs4[i++]);
      }
      break;
    } /* End Switch */
     
    if (!isLineEnd && category.isOK() && i<ucs4.size())
    {
      SS_UCS2 u2=category.encode (ucs4[i]);
      if (u2 == SD_CC_Mn || u2 == SD_CC_Me)
      {
        /*
         * There are two major differences between a shaped glyph
         * and an unshaped glyph. 
         * 1. The shaped glyph does not contain the 'composition'
         *    extra characters at the end of du4 when usePrecomposed
         *    is set and there are no decompositions. 
         *    The unshaped glyphs always contains it.
         * 2. We can not merge <base><comp1> for shapes because base 
         *    is used for shaping. We must set compIndex after base
         *    for shaped, before base for unshaped glyphs. This
         *    is just a convention here. For unshaped we don't realy
         *    care: XXXX YYYY ZZZZ or XXXX ; YYYY ZZZZ. So we use former.
         */
        if (isShaped) 
        {
          /* 1 */
          if (du4.size()==4) du4.append (composition);
          /* 2 */
          compIndex = du4.size();
        }
        else
        {
          if (du4.size()==0) /* just in case */
          {
            du4.append (composition);
            composition = 0;
            usePrecomposed = false;
            compIndex =0; /* also works with 1 - #2 is very weak */
          }
          else if (du4.size()==1) /* make this a non-composing */
          {
            composition = 0;
            usePrecomposed = false;
            compIndex =0; /* also works with 1 - #2 is very weak */
          }
          else /* composing extras */
          {
            compIndex = du4.size();
          }
        }
        while (u2 == SD_CC_Mn || u2 == SD_CC_Me)
        {
          du4.append (ucs4[i]);
          i++; 
          if (i >= ucs4.size()) break;
          u2=category.encode (ucs4[i]);
        }
      }
    }
    SGlyph g (chartype, du4, composition, mirror, 
       isShaped, clusterIndex, compIndex);
    g.usePrecomp = (usePrecomposed || du4.size()==1);
    gl->append (g);
    if (g.isEOP())
    {
      /* line can not start with pdf. */
      if (i<ucs4.size() && ucs4[i]==SGlyph::PDF) i++;
      break;
    }
  }
  return i;
}

/**
 * Get a sub paragraph.
 */
SParagraph* 
SParagraph::subParagraph(unsigned int from, unsigned int to) const
{
  /* return self */
  SParagraph* p = new SParagraph ();
  CHECK_NEW (p);
  if (from > to) return p;
  p->glyphs = glyphs;
  if (from == 0 && to == glyphs.size())
  {
    return p;
  }
  /* set the initial directionality here */
  if (to<p->glyphs.size()) p->glyphs.truncate (to);
  if (from>0) p->glyphs.remove (0, from);
  return p;
}

/**
 * put the text into visual order.
 * @param gl is a line of glyphs. 
 * FIXME: make a real bidi.
 */
void
SParagraph::toVisual ()
{
  SVector<SGlyph> in = glyphs;
  glyphs.clear();

  SBinVector<bool> stack;
  stack.append (true);
  bool lrnow = true;
  unsigned int rlends=0;

  /*
   * go through the input and construct gl.
   */
  unsigned int isize = in.size();
  for (unsigned int i=0; i<isize; i++)
  {
    if (in[i] == SGlyph::LRO)
    {
       stack.append (true);
       lrnow = true;
    }
    else if (in[i] == SGlyph::RLO)
    {
       stack.append (false);
       lrnow = false;
       for (rlends=i+1; rlends<in.size(); rlends++)
       {
          if (in[rlends] == SGlyph::LRO || in[rlends] == SGlyph::RLO
              || in[rlends] == SGlyph::PDF)
          {
             break;
          }
       }
    }
    else if (in[i] == SGlyph::PDF)
    {
       if (stack.size() > 1)
       {
         lrnow = stack[stack.size()-2];
         stack.truncate (stack.size()-1);
       }
       else
       {
         fprintf (stderr, "STextLine::insert unmatched STextType::PDF\n");
         lrnow = true;
       }
       if (!lrnow) 
       {
         for (rlends=i+1; rlends<in.size(); rlends++)
         {
           if (in[rlends] == SGlyph::LRO || in[rlends] == SGlyph::RLO
                || in[rlends] == SGlyph::PDF)
           {
              break;
           }
         }
       }
    }
    else
    {
       if (lrnow)
       {
         SGlyph g (in[i]);  
         g.lr = lrnow;
         glyphs.append (g);
       }
       else
       {
         rlends--;
         SGlyph g (in[rlends]);
         g.lr = lrnow;
         glyphs.append (g);
       }
    }
  }
}

/**
 * FIXME: apply toLogical algorithm that is not published
 * by Unicode Consortium reverse engineered by
 * Gaspar Sinai <gsinai@yudit.org> 2002-11-17
 */
void
SParagraph::toLogical()
{
  SVector<SGlyph> in = glyphs;
  glyphs.clear();

  unsigned int isize = in.size();
  SV_UCS4 d;
  SGlyph rlo (SD_CC_Xx, d, SGlyph::RLO, 0, false, 0, 0);
  SGlyph pdf (SD_CC_Xx, d, SGlyph::PDF, 0, false, 0, 0);
  for (unsigned int i=0; i<isize; i++)
  {
     const SGlyph* g = in.peek (i); 
     if (g->lr)
     {
          glyphs.append (*g);
     }
     else
     {
        unsigned int start = i;
        while (i<isize && !in[i].lr)
        {
           i++;
        }
        unsigned int end = i;
        i--; /* We will increment it later in the external loop */
        glyphs.append (rlo);
        while (end>start)
        {
          --end;
          glyphs.append (in[end]);
        }
        glyphs.append (pdf);
     }
  }
}
