/** 
 *  Yudit Unicode Editor Source File
 *
 *  GNU Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
 *  GNU Copyright (C) 2001  Gaspar Sinai <gsinai@yudit.org>  
 *  GNU Copyright (C) 2000  Gaspar Sinai <gsinai@yudit.org>  
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2,
 *  dated June 1991. See file COPYYING for details.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "stoolkit/STextData.h"

#include "stoolkit/SEncoder.h"
#include "stoolkit/SBinHashtable.h"
#include "stoolkit/SUtil.h"
#include "stoolkit/STypes.h"
#include "stoolkit/SCharClass.h"


/*------------------------------------------------------------------------------
 *                     STextDataEvent
 *------------------------------------------------------------------------------
 */
/**
 * @param attr is tru if no text was added or removed. only attribute changed.
 */
STextDataEvent::STextDataEvent (const STextIndex& _start, bool attr)
{
  start = _start;
  valid = false;
  attribute = attr;
}
/**
 * Create a stoolkit STextDataEvent
 */
STextDataEvent::STextDataEvent (void)
{
  clear ();
}

/**
 * Set the remaining.
 */
void
STextDataEvent::setRemaining (const STextIndex& _remaining)
{
  remaining = _remaining;
  valid = true;
}

/**
 * Nothing to destroy. Just in case
 */
STextDataEvent::~STextDataEvent ()
{
}

/**
 * Clear all fields.
 */
void
STextDataEvent::clear ()
{
  start = STextIndex (0,0);
  remaining =  STextIndex (0,0);
  valid = false;
  attribute = true;
}

/**
 * Add a new event to this in a sane way.
 */
void
STextDataEvent::add (const STextDataEvent& event)
{
  if (!event.valid)
  {
    return;
  }
  if (!valid)
  {
    start = event.start;
    remaining = event.remaining;
    if (!event.attribute) attribute = false;
  }
  else
  {  
    if (event.remaining < remaining)
    {
      remaining = event.remaining;
    }
    if (event.start < start)
    {
      start = event.start;
    }
  }
  valid = true;
}

/*------------------------------------------------------------------------------
 *                           STextData
 *------------------------------------------------------------------------------
 */

/**
 * Create and empty STextData 
 */
STextData::STextData (void)
{
}

/**
 * Create a text data from utf8.
 * @param utf8 is the input text
 */
STextData::STextData (const SString& utf8)
{
  insert (utf8);
}

/**
 * clear the data and assign new data
 */
STextData
STextData::operator = (const STextData & data)
{
  clear ();
  insert (data);
  return *this;
}

/**
 * Glyph Lines are pointers
 */
STextData::~STextData ()
{
  for (unsigned int i=0; i<lines.size(); i++)
  {
     delete lines[i];
  }
}

/**
 * Get the contained text
 */
SString
STextData::getText () const
{
  if (size()==0) return SString();
  return getText (STextIndex (0, 0), STextIndex (size(), size(size()))); 
}

/**
 * Get the text between the cursor and here
 */
SString
STextData::getText (const STextIndex& index) const
{
  return getText (textIndex, index); 
}

/**
 * Extract a bi-di marked utf-8 text.
 * @param begin is the starting index
 * @param end is the ending index
 */
SString
STextData::getText (const STextIndex& begin, const STextIndex& end) const
{
  SString ret;
  if (size() == 0) return SString(ret);

  STextIndex start;
  STextIndex stop;
  if (end > begin)
  {
    start = begin;
    stop = end;
  }
  else
  {
    start = end;
    stop = begin;
  }
  if (stop.line > size())
  {
     stop.line = size(); stop.index = size(size()-1);
  }
  for (unsigned int i=start.line; i<=stop.line && i<size(); i++)
  {
     unsigned int b = (i==start.line) ? start.index : 0;
     unsigned int e = (i==stop.line)? stop.index : size (i);
     ret.append (getText (i, b, e));
  }
  return SString(ret);
}

/**
 * Extract a bi-di marked utf-8 text.
 * TODO: implements a better bidi.
 * @param beginLine is the starting line
 * @param beginIndex is the starting position
 * @param endIndex is the ending position
 */
SString
STextData::getText (unsigned int beginLine, unsigned int beginIndex,
   unsigned int endIndex) const
{
  SV_UCS4 ucs;

  const SParagraph* line = lines[beginLine];
  SParagraph* sub =  line->subParagraph (beginIndex, endIndex);
  sub->toLogical();
  for (unsigned int j=0; j<sub->size(); j++)
  {
    ucs.append ((sub->peek(j))->getChars());
  }
  delete sub;
  SEncoder ic("utf-8-s");
  return (SString(ic.encode (ucs)));
}

/**
 * find string in text data from current position.
 * return an index to the end of that data.
 * move the current position to the beginning of the data.
 * it does not search across lines.
 */
STextIndex
STextData::find (const SString& string)
{
  STextData d (string);

  /**
   * go through all lines.
   */
  for (unsigned int i=textIndex.line; i<size(); i++)
  {
     unsigned int begin = (i==textIndex.line) ? textIndex.index : 0;
     unsigned int end = size (i);

     SParagraph* line = lines[i];
     /**
      * go through all glyphs
      */
     for (unsigned int j=begin; j+d.size(0)<=end; j++)
     {
        /* hack. We don't disclose the SGlyphLIne= */
        bool found = true;
        for (unsigned int k=0; k<d.size(0); k++)
        {
          const SGlyph* g = line->peek(j+k);
          const SGlyph& g2 = d.glyphAt (STextIndex(0, k));
          if (g2 != *g)
          {
            found = false;
          }
        }
        if (found)
        {
          move (STextIndex(i, j));
          return STextIndex (i, j+d.size(0));
        }
     }
  }
  return STextIndex (0,0);
}

/**
 * Get a glyph at a certain position
 * @param index is the glyph index
 * @return a reference to the glyph.
 */
const SGlyph&
STextData::glyphAt (const STextIndex& index) const
{
  return (*lines[index.line])[index.index];
}

/**
 * move the insertion point.
 * @param line is the line number.
 * @param index is the line index.
 */
void
STextData::move (const STextIndex& index)
{
  STextIndex ndx (index);
  if (ndx.line > size()) ndx.line = size();
  if (ndx.line > 0 && !isProperLine (ndx.line-1))
  {
     ndx.line--;
     ndx.index = size(ndx.line);
  }
  if (ndx.line == size())
  {
    ndx.index = 0;
  }
  else if (ndx.index >= size(ndx.line))
  {
    ndx.index = size(ndx.line);
    if (isProperLine (ndx.line) && ndx.index == size(ndx.line))
    {
      ndx.index = ndx.index-1;
    }
  }
  textIndex = ndx;
}

/**
 * set the text data
 */
void
STextData::setText (const SString& string)
{
  SEncoder ic("utf-8-s");
  SV_UCS4 ucs4 = ic.decode (string, false);
  setText (ucs4);
}

/**
 * set the text data
 */
void
STextData::setText (const SV_UCS4& ucs4)
{
  clear ();
  /* split into lines */
  unsigned int start = 0;
  while (true)
  {
    SParagraph* gl = new SParagraph(ucs4, &start);
    if (gl->size()==0) 
    {
      delete gl; 
      break;
    }
    gl->toVisual();
    /* shaping is done on the visual buffer here. */
    gl->reShapeSegment (0, 0);
    lines.append (gl);
  } 
  STextDataEvent nevt (STextIndex(0,0));
  event.add (nevt);
}

/**
 * insert one signle glyph and move the cursor to the right
 * @param glyph is the glyph to insert
 * @return an index to end, text between current cursor and end
 * has been inserted.
 */
void
STextData::insert (const SGlyph& glyph, STextIndex* ret)
{
  STextIndex oldIndex = textIndex;

  if (size() <= textIndex.line)
  {
     lines.append (new SParagraph());
  }

  SParagraph* line = lines[textIndex.line];
  line->insert (textIndex.index, glyph);
  if (ret->line == textIndex.line)
  {
    ret->index++;
  }
  textIndex.index = textIndex.index+1;

  STextDataEvent nevt (oldIndex);
  /* there are two cases. we opened a new line or not */
  if (glyph.isEOP())
  {
    /* cut line in two */
    if (line->size() > textIndex.index)
    {
      SParagraph *newLine = new SParagraph (*line);
      newLine->remove (0, textIndex.index);
      line->truncate (textIndex.index);
      lines.insert (textIndex.line+1, newLine);

      textIndex.index = newLine->size()-1;
      textIndex.line = textIndex.line + 1;
      setMaxLimits (&nevt, textIndex);
      textIndex.index = 0;
      ret->line++;
      if (ret->line== textIndex.line) ret->index=0;
    }
    else /* new line */
    {
      textIndex.index = 0;
      textIndex.line = textIndex.line + 1;
      setMaxLimits (&nevt, textIndex);
      ret->line++;
      ret->index=0;
      if (ret->line== textIndex.line) ret->index=0;
    }
  }
  else
  {
    setMaxLimits (&nevt, textIndex);
  }
  event.add (nevt);
}

/**
 * Insert a text from data
 * @param data is the STextData
 */
void
STextData::insert (const STextData& data)
{
  STextIndex ind;
  for (unsigned int i=0; i<data.size(); i++)
  {
    for (unsigned int j=0; j<data.size(i); j++)
    {
      insert (data.glyphAt (STextIndex (i, j)), &ind);
    }
  }
}

/**
 * Insert a text from utf8. Do fuzzy algoritm to determine
 * direction and all kinds of fuzzy properties, like composition.
 * The resulting text is added to the textbuffer in screen-order.
 * FIXME: 
 * - Implement a bidi algoritm. The current version treats everything
 *   as strong left-to-right character unless it is enclosed in  RLO-PDF.
 * - Get a table of all composing characters and break down the glyphs so
 *   that one glyph contains the base character and the compositng marks.
 * @param utf8 an utf8 encoded text.
 * @return an index to end, text between current cursor and end
 * has been inserted. 
 */
STextIndex
STextData::insert (const SString& utf8)
{
  SEncoder ic("utf-8-s");
  SV_UCS4 ucs4 = ic.decode (utf8, false);
  /* split into lines */
  unsigned int start = 0;
  STextIndex old=textIndex;
  unsigned int i;
  while (true)
  {
    SParagraph gl (ucs4, &start);
    if (gl.size() ==0) break;

    gl.toVisual();

    /* we are all visual */
    const SGlyph* before = 0;
    const SGlyph* after = 0;

    STextIndex index = textIndex;
    if (index.line < size())
    {
      SParagraph *line = lines[index.line];; 
      /* full line before=0, after=0*/
      int cbefore = line->getNonTransparentBefore (index.index+1);
      int cafter = line->getNonTransparentAfter (index.index);
      if (cbefore >= 0) before = line->peek((unsigned int)cbefore);
      if (cafter >= 0) after = line->peek((unsigned int)cafter);
    }
    gl.reShapeSegment (before, after);

    for (i=0; i<gl.size(); i++)
    {
      insert (gl[i], &old);
    }
  } 

  /* shape the begining and ending position */
  STextIndex vold = getTextIndex (old, -1); 
  reShape (vold);

  /* before */
  STextIndex op = getTextIndex (vold, -1);
  if (vold != op) reShape (op);

  /* after */
  STextIndex oa = getTextIndex (vold, 1);
  if (vold != oa) reShape (oa);

  reShape (textIndex);
  /* before */
  STextIndex tp = getTextIndex (textIndex, -1);
  if (textIndex != tp) reShape (tp);

  /* before */
  STextIndex ta = getTextIndex (textIndex, 1);
  if (textIndex != ta) reShape (ta);

  return STextIndex (old);
}

/**
 * remove glyphs between current and line and size.
 * current can be less or greater.
 * @param index is the ending high end is non-exclusive
 */
void
STextData::remove (const STextIndex& index)
{
  if (size() == 0) return;

  STextIndex newIndex = reorder (index);
  unsigned int i;
  bool join = false;
  for (i=textIndex.line; i<=newIndex.line && i<size(); i++)
  {
     unsigned int begin = (i==textIndex.line) ? textIndex.index : 0;
     unsigned int end = (i==newIndex.line)? newIndex.index : size (i);

     SParagraph* line = lines[i];
     if (begin == 0 && end == size(i))
     {
        if (isProperLine(i)) join = true;
        line->clear();
     }
     else for (unsigned int j=begin; j<end; j++)
     {
        /* hack. We don't disclose the SGlyphLIne= */
        if (j+1==end)
        {
          SGlyph g = glyphAt (STextIndex(i, begin));
          if (g.isEOP()) join = true;
        }
        line->remove (begin);
     }
  }
  unsigned int lr = 0;

  /* Remove zeroes */
  for (i=textIndex.line+1; i<=newIndex.line && i<size() + lr; i++)
  {
     SParagraph* line = lines[i-lr];
     if (line->size() == 0)
     {
       lines.remove (i-lr);
       delete line;
       lr++;
     }
  }

  /* now we have everything collapsed. */

  STextDataEvent nevt (textIndex);
  setMaxLimits (&nevt, textIndex);
  event.add (nevt);

  /* a joining line was created */
  if (join)
  {
    /* wanted, but impossible */
    if (textIndex.line+1 == size())
    {
      SParagraph* line = lines[textIndex.line];
      if (line->size()==0)
      {
        lines.remove (textIndex.line);
        delete line;
      }
      return;
    }
    SParagraph* gl = lines[textIndex.line+1];
    lines.remove (textIndex.line+1);

    STextIndex sv = textIndex;
    STextIndex dm = textIndex;
    for (i=0; i<gl->size(); i++)
    {
      insert ((*gl)[i], &dm);
    }
    delete gl;
    move (sv);
  }
  else
  { /* we may not removed this */
    SParagraph* line = lines[textIndex.line];
    if (line->size()==0)
    {
      lines.remove (textIndex.line);
      delete line;
    }
  }
  reShape (textIndex);
  /* before */
  STextIndex tp = getTextIndex (textIndex, -1);
  if (textIndex != tp) reShape (tp);

  /* before */
  STextIndex ta = getTextIndex (textIndex, 1);
  if (textIndex != ta) reShape (ta);
}

/**
 * Modify this glyph by adding extra composing characters
 * to it. 
 * @param c is a new composing character to be added.
 * @param toleft is true if we need to add it to the left of cursor.
 * @return true if this was a composing character.
 */
bool
STextData::addComposing(SS_UCS4 c, bool toleft)
{
  if (textIndex.line >= lines.size())
  {
     return false;
  }
  SParagraph* line = lines[textIndex.line];
  SGlyph* g = 0;
  STextIndex fromIndex = textIndex;
  STextIndex toIndex = textIndex;

  if (toleft && textIndex.index > 0 && textIndex.index-1 < line->size())
  {
    g = (SGlyph*) line->peek(textIndex.index-1);
    fromIndex.index = fromIndex.index-1;
  }
  if (!toleft && textIndex.index < line->size() && line->size() > 0)
  {
    g = (SGlyph*) line->peek(textIndex.index);
    toIndex.index = toIndex.index+1;
  }
  /* Hack - this is read only */
  bool status = g ? g->addComposing(c) : false;
  if (status)
  {
    reShape (fromIndex);
    reShape (toIndex);
    STextDataEvent nevt (fromIndex, false);
    setMaxLimits (&nevt, toIndex);
    event.add (nevt);
  }
  return status;
}

/**
 * @return the removed composing character if any
 * @param toleft is true if we need to remove it from the left of cursor.
 * return 0 if there are no more composing characters.
 */
SS_UCS4
STextData::removeComposing(bool fromleft)
{
  if (textIndex.line >= lines.size())  return false;
  SParagraph* line = lines[textIndex.line];
  SGlyph* g = 0;
  STextIndex fromIndex = textIndex;
  STextIndex toIndex = textIndex;
  if (fromleft && textIndex.index > 0 && textIndex.index-1 < line->size())
  {
    g = (SGlyph*) line->peek(textIndex.index-1);
    fromIndex.index = fromIndex.index-1;
  }
  if (!fromleft && textIndex.index < line->size() && line->size() > 0)
  {
    g = (SGlyph*) line->peek(textIndex.index);
    toIndex.index = toIndex.index+1;
  }
  /* Hack - this is read only */
  SS_UCS4 status = g ? g->removeComposing() : false;
  if (status)
  {
    reShape (fromIndex);
    reShape (toIndex);
    STextDataEvent nevt (fromIndex, false);
    setMaxLimits (&nevt, toIndex);
    event.add (nevt);
  }
  return status;
}
/**
 * select a region of text and move cursor to it.
 * @param index is the new index. high end is non-exclusive
 * @param is is true if select, unselect otherwise. 
 */
void
STextData::select (const STextIndex& index, bool is)
{
  if (size() == 0) return;

  STextIndex newIndex = reorder (index);
  bool reordered = (newIndex != index);


  if (newIndex.line > size())
  {
     newIndex.line = size(); newIndex.index = size(size()-1);
  }
  for (unsigned int i=textIndex.line; i<=newIndex.line && i<size(); i++)
  {
     unsigned int begin = (i==textIndex.line) ? textIndex.index : 0;
     unsigned int end = (i==newIndex.line)? newIndex.index : size (i);

     SParagraph* line = lines[i];
     if (end > line->size()) end = line->size(); 
     for (unsigned int j=begin; j<end; j++)
     {
        /* hack. We don't disclose the SGlyphLIne= */
        SGlyph* g = (SGlyph*) line->peek(j);
        g->selected = is;
     }
  }
  if (reordered)
  {
    STextDataEvent nevt (textIndex, true);
    setMaxLimits (&nevt, newIndex);
    event.add (nevt);
  }
  else
  {
    STextDataEvent nevt (textIndex, true);
    setMaxLimits (&nevt, newIndex);
    event.add (nevt);
    textIndex =  newIndex;
  }
}

bool
STextData::setLineType (const SString& str)
{
  if (str==SS_LB_UNIX) return setLineType (SGlyph::ELF);
  if (str==SS_LB_MAC) return setLineType (SGlyph::ECR);
  if (str==SS_LB_DOS) return setLineType (SGlyph::ECRLF);
  if (str==SS_LB_LS) return setLineType (SGlyph::ELS);
  if (str==SS_LB_PS) return setLineType (SGlyph::EPS);
  return false;
}

/**
 * Set the line type for the whole data structure.
 * @param type will be forced after each newline. 
 * @return true if anything changed.
 */
bool
STextData::setLineType (SGlyph::SLineEnd type)
{
  SV_UCS4 v;
  switch (type)
  {
  case SGlyph::ELF:
    v.append ((SS_UCS2) SGlyph::LF);
    return setLineType (SGlyph(SD_CC_Cc, v, (SS_UCS4) SGlyph::LF, 0, false, 0, 0));
  case SGlyph::ECR:
    v.append ((SS_UCS2) SGlyph::CR);
    return setLineType (SGlyph(SD_CC_Cc, v, (SS_UCS4) SGlyph::CR, 0, false, 0, 0));
  case SGlyph::ECRLF:
    v.append ((SS_UCS2) SGlyph::CR);
    v.append ((SS_UCS2) SGlyph::LF);
    return setLineType (SGlyph(SD_CC_Cc, v, 0, 0, false, 0, 0));
  case SGlyph::ELS:
    v.append ((SS_UCS2) SGlyph::LS);
    return setLineType (SGlyph(SD_CC_Zl, v, (SS_UCS4) SGlyph::LS, 0, false, 0, 0));
  case SGlyph::EPS:
    v.append ((SS_UCS2) SGlyph::PS);
    return setLineType (SGlyph(SD_CC_Zp, v, (SS_UCS4) SGlyph::PS, 0, false, 0, 0));
  }
  return false;
}
/**
 * Set the line type for the whole data structure.
 * @param type will be forced after each newline. 
 * @return true if anything changed.
 */
bool
STextData::setLineType (const SGlyph& _glyph)
{
  bool retvle = false;
  SGlyph glyph = _glyph;
  glyph.lr = true;
  for (unsigned int i=0; i<lines.size(); i++)
  {
    SParagraph* gl = lines[i];
    if (gl==0 || gl->size()==0) continue;
    const SGlyph& g = *gl->peek(gl->size()-1);
    if (!g.isEOP())
    {
      continue;
    }
    //if (!g.selected) continue;
    if (glyph == g)
    {
      continue;
    }
    glyph.selected = g.selected;
    gl->replace (gl->size()-1, glyph);
    retvle = true;
  }
  return retvle;
}

/**
 * select a region of text and move cursor to it.
 * @param index is the new index - high end is non-exclusive
 * @param is is true if underline, un-underline otherwise.
 */
void
STextData::underline (const STextIndex& index, bool is)
{
  if (size() == 0) return;

  STextIndex newIndex = reorder (index);
  bool reordered = (newIndex != index);

  for (unsigned int i=textIndex.line; i<=newIndex.line && i<size(); i++)
  {
     unsigned int begin = (i==textIndex.line) ? textIndex.index : 0;
     unsigned int end = (i==newIndex.line)? newIndex.index : size (i);

     SParagraph* line = lines[i];
     if (end > line->size()) end = line->size(); 
     for (unsigned int j=begin; j<end; j++)
     {
        /* hack. We don't disclose the SGlyphLIne= */
        SGlyph* g = (SGlyph*) line->peek(j);
        g->underlined = is;
     }
  }
  if (reordered)
  {
    STextDataEvent nevt (textIndex, true);
    setMaxLimits (&nevt, newIndex);
    event.add (nevt);
  }
  else
  {
    STextDataEvent nevt (textIndex, true);
    setMaxLimits (&nevt, newIndex);
    event.add (nevt);
    textIndex =  newIndex;
  }
}

/**
 * Clear the text
 */
void
STextData::clear ()
{
  for (unsigned int i=0; i<lines.size(); i++)
  {
     delete lines[i];
  }
  lines.clear();
  textIndex = STextIndex (0, 0);
  STextDataEvent nevt (STextIndex (0, 0));
  nevt.remaining = STextIndex (0, 0);
  nevt.valid = true;
  event.add (nevt);
}

void
STextData::fireEvent ()
{
  if (!event.valid || listener == 0) return ;
  listener->textChanged (this, event); 
  event.clear();
}

void
STextData::clearEvent ()
{
  event.clear();
}

/**
 * return the number of lines in text
 */
unsigned int
STextData::size() const
{
  return lines.size();
}

/**
 * return the number of lines in line
 */
unsigned int
STextData::size(unsigned int line) const
{
  if (line >= size()) return 0;
  const SParagraph* l = lines[line];
  return l->size();
}

/**
 * Get the text index relative to offset.
 * @param offset is character offset to textIndex.
 */
STextIndex
STextData::getTextIndex(int charOffset) const
{
  return getTextIndex (textIndex, charOffset);
}
/**
 * Get the text index relative to offset.
 * @param offset is character offset to textIndex.
 */
STextIndex
STextData::getTextIndex(const STextIndex& base, int charOffset) const
{
  STextIndex ret =  base;
  if (charOffset > 0)
  {
    for (int i=0; i<charOffset; i++)
    {
      if (ret == STextIndex (size(), 0)) break;
      if (size(ret.line) <= ret.index+1)
      {
        if (!isProperLine (ret.line))
        {
          if (size(ret.line) == ret.index+1)
          {
             ret.index = ret.index + 1;
          }
          break;
        }
        ret.index = 0;
        ret.line = ret.line + 1;
        continue;
      }
      ret.index = ret.index + 1;
      continue;
    }
  }
  else if (charOffset < 0)
  {
    for (int i=0; i>charOffset; i--)
    {
      if (ret == STextIndex (0, 0)) break;
      if (ret.index > 0)
      {
        ret.index = ret.index - 1;
        continue;
      }
      ret.line = ret.line-1;
      /*
      * For proper lines,
      * The beginning of the line is the same location as the end. 
      */
      ret.index = size(ret.line);
      if (isProperLine (ret.line))
      {
        ret.index = ret.index-1;
      }
    }
     
  }
  return STextIndex (ret);
}

/**
 * Check if line ends with newline glyph
 */
bool
STextData::isProperLine (unsigned int line) const
{
  if (line >= size()) return false;
  if (size(line)==0) return false;
  const SGlyph g = glyphAt(STextIndex(line, size(line)-1));
  return g.isEOP();
}

/**
 * add data listener.
 * TODO: now it only sets it.
 */
void
STextData::addTextDataListener (STextDataListener* _listener)
{
  listener = _listener;
}

/**
 * Set the maximum limits.
 * @param evt -> remainingLine and remainingPosition will be set
 * @param index is the one that needs to be converted to remaining index
 */
void
STextData::setMaxLimits (STextDataEvent* evt, const STextIndex& index)
{
  evt->valid = true;
  unsigned int lsize = size();
  if (index.line >= lsize)
  {
     evt->setRemaining (STextIndex (0,0));
     return;
  }
  unsigned int rem  = lsize - index.line - 1;
  unsigned int cindex = size(index.line);
  evt->setRemaining (STextIndex (rem, (index.index>=cindex) ? 0 : cindex - index.index));
}

/**
 * Convert remainin-line to maxline
 */
STextIndex
STextData::getMaxTextIndex (const STextDataEvent& evt) const
{
  if (!evt.valid) return STextIndex (0,0);
  unsigned int lsize = size();
  if (lsize <= evt.remaining.line)
  {
     return STextIndex (lsize,0);
  }
  unsigned int line = lsize - evt.remaining.line -1;
  if (size(line) < evt.remaining.index)
  {
     return STextIndex (line, size(line));
  }
  return STextIndex (line, size(line)+1-evt.remaining.index);
}

/**
 * Put the stuff in order, so that it will always be in increasing order.
 * @param end is the desired end
 */
STextIndex
STextData::reorder (const STextIndex& index)
{
  if (index > textIndex)
  {
     return STextIndex (index);
  }
  STextIndex tmp = textIndex;
  textIndex = index;
  return STextIndex (tmp);
}  

/**
 * return true if character is a whitespace
 */
bool
STextData::isWhiteSpace (const STextIndex& index) const
{
  return glyphAt (index).isWhiteSpace();
}


/**
 * return true if character is a number. Addition by Maarten van Gompel <proycon@anaproy.homeip.net>
 */
bool
STextData::isNumber (const STextIndex& index) const
{
  return glyphAt (index).isNumber();
}


/**
 * return true if character is a target for select. Addition by Maarten van Gompel <proycon@anaproy.homeip.net>
 */
bool
STextData::isDelimiter (const STextIndex& index) const
{
  return glyphAt (index).isDelimiter();
}

/**
 * Re-shape current glyph.
 * The full line is taken!
 * @param index is the index around which we re-shape.
 */
void
STextData::reShapeOne(const STextIndex& index)
{
  /* current */
  if (index.line < size() && index.index < size (index.line))
  {
    SParagraph *line = lines[index.line];; 

    /* full line before=0, after=0*/
    const SGlyph* before = 0;
    const SGlyph* after = 0;
    /* This is the insertion point! */
    int cbefore = line->getNonTransparentBefore (index.index);
    int cafter = line->getNonTransparentAfter (index.index);
    if (cbefore >= 0) before = line->peek((unsigned int)cbefore);
    if (cafter >= 0) after = line->peek((unsigned int)cafter);
    SGlyph* g = (SGlyph*) line->peek(index.index);

    // Hack for speed. This is a constant in reality ...
    if (g->setShape(before, after))
    {
       /* this glyph changed shape */
       STextDataEvent nevt (index, true);
       setMaxLimits (&nevt, index);
       event.add (nevt);
    }
  }
}

/**
 * Re-shape current glyph.
 * The full line is taken!
 * @param index is the index around which we re-shape.
 */
void
STextData::reShape(const STextIndex& index)
{
  /* current */
  if (index.line < size() && index.index < size (index.line))
  {
    SParagraph *line = lines[index.line];; 

    /* full line before=0, after=0*/
    const SGlyph* before = 0;
    const SGlyph* after = 0;
    int cbefore = line->getNonTransparentBefore (index.index);
    int cafter = line->getNonTransparentAfter (index.index);
    if (cbefore >= 0) before = line->peek((unsigned int)cbefore);
    if (cafter >= 0) after = line->peek((unsigned int)cafter);
    SGlyph* g = (SGlyph*) line->peek(index.index);

    // Hack for speed. This is a constant in reality ...
    if (g->setShape(before, after))
    {
       /* this glyph changed shape */
       STextDataEvent nevt (index, true);
       setMaxLimits (&nevt, index);
       event.add (nevt);
    }
    /* do it after and before to skip spaces */
    if (cbefore>=0)
    {
       STextIndex ib = index;
       ib.index = (unsigned int) cbefore;
       reShapeOne (ib);
    }
    if (cafter>=0)
    {
       STextIndex ia = index;
       ia.index = (unsigned int) cafter;
       reShapeOne (ia);
    }
  }
}

